package me.unurled.raxen.utils;

import com.google.gson.Gson;
import de.tr7zw.nbtapi.NBTItem;
import me.unurled.raxen.Raxen;
import me.unurled.raxen.components.items.NBT;
import net.kyori.adventure.text.Component;
import org.bukkit.Bukkit;
import org.bukkit.Material;
import org.bukkit.enchantments.Enchantment;
import org.bukkit.inventory.Inventory;
import org.bukkit.inventory.ItemFlag;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.PlayerInventory;
import org.bukkit.inventory.meta.ItemMeta;
import org.bukkit.util.io.BukkitObjectInputStream;
import org.bukkit.util.io.BukkitObjectOutputStream;
import org.jetbrains.annotations.NotNull;
import org.yaml.snakeyaml.external.biz.base64Coder.Base64Coder;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.*;

import static me.unurled.raxen.utils.Utils.*;

public class Items {

    /**
     * create an item with @param
     * @param material the item material
     * @param amount the amount of the itemstack
     * @param glow if the item glows or not
     * @param unb if the item is unbreakable
     * @param name the name of the item
     * @param lore the lore of the item
     * @return the created itemstack
     */
    public static @NotNull ItemStack createItem(Material material, int amount, boolean glow, boolean unb, String name, String... lore) {
        ItemStack item = new ItemStack(material, amount);
        ItemMeta itemMeta = item.getItemMeta();
        if(name != null) {
            itemMeta.displayName(colorComp(name));
        }
        if(lore != null) {
            List<Component> list = new ArrayList<>();
            for(String string :  lore) {
                list.add(colorComp(string));
            }
            itemMeta.lore(list);
        }
        if(glow) {
            itemMeta.addItemFlags(ItemFlag.HIDE_ENCHANTS);
            itemMeta.addEnchant(Enchantment.DURABILITY, 1, true);
        }
        if(unb) {
            itemMeta.setUnbreakable(true);
        }
        item.setItemMeta(itemMeta);
        return item;
    }

    /**
     * edit an item with provided @param
     * @param item the item to modify
     * @param name the new name of the item
     * @param amount the new amount
     * @param lore the new lore
     * @return the modified itemstack
     */
    public static @NotNull ItemStack editItem(ItemStack item, String name, int amount, List<Component> lore) {
        if(amount == 0) {
            amount = 1;
        }
        item.setAmount(amount);
        ItemMeta itemMeta = item.getItemMeta();
        if (name != null) {
            itemMeta.displayName(Component.text(name));
        }
        itemMeta.lore(lore);
        item.setItemMeta(itemMeta);
        return item;
    }

    /**
     * the basic close gui button
     * @return a barrier block itemstack named close
     */
    public static @NotNull ItemStack closeItem () {
        ItemStack close = createItem(Material.BARRIER, 1, false, false, colorString("<red>CLOSE</red>"));
        ItemMeta clo = close.getItemMeta();
        clo.setCustomModelData(1);
        close.setItemMeta(clo);
        return close;
    }

    /**
     * fill item in chest and guis
     * @return a grey glass pane named nothing
     */
    public static @NotNull ItemStack greyPane() {
        return createItem(Material.GRAY_STAINED_GLASS_PANE, 1, false, true, "");
    }

    /**
     * serialize item
     * @param it an itemstack
     * @param slot the slot it was in the inventory
     * @return the serialized string
     */
    public static @NotNull String itemStackSerialize(ItemStack it, @NotNull Integer slot) {
        String str = "";
        str = itemTo64(it);
        str = str + "@#SLOT#" + (slot-1);
        return str;
    }

    /**
     * Serelize the PlayerInventory
     * @param inv a player inventory
     * @return return a serelized string
     */
    public static @NotNull String itemStackSer(PlayerInventory inv) {
        List<String> list = new ArrayList<>();
        Integer i = 0;
        for (ItemStack it : inv.getContents()) {
            String str = "";
            str = itemTo64(it);
            if (it != null) {
                if (isContained(it, inv.getArmorContents())) {
                    str = str + "@#SLOT#" + getWhichArmor(it.getType());
                } else {
                    i++;
                    str = str + "@#SLOT#" + (i - 1);
                }
            }
            list.add(str);
        }
        return listItemStackSerialize(list);
    }

    /**
     * if item is in its
     * @param item an Itemstack
     * @param its an Itemstack[]
     * @return return if item is in its
     */
    public static boolean isContained(ItemStack item, ItemStack[] its) {
        for (ItemStack it : its) {
            if (it == item) {
                return true;
            }
        }
        return false;
    }

    /**
     * Get which armor the type is
     * @param type a material type
     * @return the armor type or null
     */
    public static String getWhichArmor(Material type) {
        String name = type.name();
        if (name.contains("HELMET") || name.contains("helmet")) {
            return "HELMET";
        } else if (name.contains("CHESTPLATE") || name.contains("chestplate")) {
            return "CHESTPLATE";
        } else if (name.contains("LEGGINGS") || name.contains("leggins")) {
            return "LEGGINGS";
        } else if (name.contains("BOOTS") || name.contains("boots")) {
            return "BOOTS";
        }
        return null;
    }

    /**
     * a class that represent the itemstack deserialized.
     */
    public static class itemStackDeserializeResult {
        int slot;
        ItemStack it;
        String armor;

        /**
         * custructor
         * @param slot an int of the slot of the item
         * @param it the item which is deserialized.
         */
        public itemStackDeserializeResult(Integer slot, ItemStack it) {
            this.slot = slot;
            this.it = it;
        }

        public itemStackDeserializeResult(Integer slot, ItemStack it, String armor) {
            this.slot = slot;
            this.it = it;
            this.armor = armor;
        }
    }

    /**
     * a class that represent a list of itemstack deserialized.
     */
    public static class itemStackListDeserilize  {
        HashMap<Integer, ItemStack> inv;

        /**
         * constructor
         * @param inv a list of itemstack with their slot.
         */
        public itemStackListDeserilize(HashMap<Integer, ItemStack> inv) {
            this.inv = inv;
        }
    }

    /**
     * deserilize items
     * @param str the serilized item
     * @return a deserialized item instance
     */
    public static @NotNull Items.itemStackDeserializeResult itemStackDeserilize(@NotNull String str)  {
        List<String> mapp = Arrays.asList(str.split("@#SLOT#"));
        Gson gson = new Gson();
        ItemStack it = new ItemStack(Material.AIR);
        Integer slot = 0;
        String armor = null;
        debug("full item " + str);
        for (String s : mapp) {
            if(s != null) {
                try {
                    it = itemFrom64(s);
                }   catch (Exception e) {
                    if (!s.equals("")) {
                        debug("slot " + s);
                        try {
                            slot = Integer.parseInt(s);
                        } catch (NumberFormatException nfe) {
                            armor = s;
                        }
                    }
                }

            }
        }
        itemStackDeserializeResult itt;
        if (armor != null) {
            itt  = new itemStackDeserializeResult(slot, it, armor);
        } else {
            itt  = new itemStackDeserializeResult(slot, it);
        }
        return itt;
    }

    /**
     * a list of serialized item in one string
     * @param list the list of all serialized item
     * @return a string
     */
    public static String listItemStackSerialize(@NotNull List<String> list) {
        Gson gson = new Gson();
        String str = "";
        for(String a : list) {
            str = str + "@#NEW_ITEM#" + a;
        }
        return str;
    }

    /**
     * deserialize item
     * @param str a list of all serialized item
     * @return a list of itemstack with their slot
     */
    public static @NotNull HashMap<Integer, ItemStack> mapItemStackDeserialize(@NotNull String str) {
        Gson gson = new Gson();
        List<String> map = Arrays.asList(str.split("@#NEW_ITEM#"));
        HashMap<Integer, ItemStack> inv = new HashMap<>();
        ArrayList<ItemStack> items = new ArrayList<ItemStack>();
        for(String s : map) {
            ItemStack itt;
            Integer slot;
            if(s != null) {
                itemStackDeserializeResult itm = itemStackDeserilize(s);
                itt = itm.it;
                if (itt != null && itt.getType() != Material.AIR) {
                    items.add(itt);
                }
                slot = itm.slot;
                inv.put(slot, itt);
            }
        }
        ItemStack[] it = items.toArray(new ItemStack[0]);
        return inv;
    }

    /**
     * deserialize item
     * @param str a list of all serialized item
     * @return a list of itemstack with their slot
     */
    public static @NotNull List<itemStackDeserializeResult> listItemStackDeserialize(@NotNull String str) {
        Gson gson = new Gson();
        List<String> map = Arrays.asList(str.split("@#NEW_ITEM#"));
        List<itemStackDeserializeResult> inv = new ArrayList<>();
        ArrayList<ItemStack> items = new ArrayList<ItemStack>();
        for(String s : map) {
            ItemStack itt;
            Integer slot;
            if(s != null) {
                itemStackDeserializeResult itm = itemStackDeserilize(s);
                itt = itm.it;
                if (itt != null && itt.getType() != Material.AIR) {
                    items.add(itt);
                }
            }
        }
        return inv;
    }

    /**
     * set all item in the list to their slot in inv
     * @param list a list of item with their slot
     * @param inv the destined inventory
     * @return the inv
     */
    public static Inventory setItemsToInventory(@NotNull HashMap<Integer, ItemStack> list, Inventory inv) {
        for (Map.Entry<Integer, ItemStack> entry : list.entrySet()) {
            inv.setItem(entry.getKey(), entry.getValue());
        }
        return inv;
    }

    /**
     * set all item in the list to their slot in inv
     * @param list a list of item with their slot
     * @param inv the destined inventory
     * @return the inv
     */
    public static PlayerInventory setItemsToInventory(@NotNull List<itemStackDeserializeResult> list, PlayerInventory inv) {
        Inventory invv = Bukkit.createInventory(null, 54);
        for (itemStackDeserializeResult itm : list) {
            if (itm.armor != null) {
                setArmorItem(inv, itm.it);
            } else {
                inv.setItem(itm.slot, itm.it);
            }
        }
        return inv;
    }

    public static void setArmorItem(PlayerInventory inv, ItemStack item) {
        String name = item.getType().name();
        if (name.contains("HELMET")) {
            inv.setHelmet(item);
        } else if (name.contains("CHESTPLATE")) {
            inv.setChestplate(item);
        } else if (name.contains("LEGGINGS")) {
            inv.setLeggings(item);
        } else if (name.contains("BOOTS")) {
            inv.setBoots(item);
        }
    }

    /**
     * set all item in the list to their slot in inv
     * @param list a list of item with their slot
     * @param inv the player inventory
     * @return the inv
     */
    public static PlayerInventory setItemsToInventory(@NotNull HashMap<Integer, ItemStack> list, PlayerInventory inv) {
        for (Map.Entry<Integer, ItemStack> entry : list.entrySet()) {
            if (entry.getKey() == 40) {
                inv.setHelmet(entry.getValue());
            } else if (entry.getKey() == 39) {
                inv.setChestplate(entry.getValue());
            } else if (entry.getKey() == 38) {
                inv.setLeggings(entry.getValue());
            } else if (entry.getKey() == 37) {
                inv.setBoots(entry.getValue());
            } else if (entry.getKey() == 41) {
                inv.setItemInOffHand(entry.getValue());
            }
            else  {
                inv.setItem(entry.getKey(), entry.getValue());
            }
        }
        return inv;
    }

    /**
     * serialize an item
     * @param stack the item
     * @return a string
     * @throws IllegalStateException stream errors
     */
    private static @NotNull String itemTo64(ItemStack stack) throws IllegalStateException {
        try {
            ByteArrayOutputStream outputStream = new ByteArrayOutputStream();
            BukkitObjectOutputStream dataOutput = new BukkitObjectOutputStream(outputStream);
            dataOutput.writeObject(stack);

            // Serialize that array
            dataOutput.close();
            return Base64Coder.encodeLines(outputStream.toByteArray());
        }
        catch (Exception e) {
            throw new IllegalStateException("Unable to save item stack.", e);
        }
    }

    /**
     * deserialize item for string to itemstack
     * @param data string
     * @return itemstack
     * @throws IOException stream error
     */
    private static ItemStack itemFrom64(String data) throws IOException {
        try {
            ByteArrayInputStream inputStream = new ByteArrayInputStream(Base64Coder.decodeLines(data));
            BukkitObjectInputStream dataInput = new BukkitObjectInputStream(inputStream);
            try {
                return (ItemStack) dataInput.readObject();
            } finally {
                dataInput.close();
            }
        }
        catch (ClassNotFoundException e) {
            throw new IOException("Unable to decode class types.", e);
        }
    }

    /**
     * return in correct naming each attributes,
     * @param str the nbt attribute
     * @return the good name
     */
    @Deprecated
    public static @NotNull String attributes(@NotNull String str) {
        String stt = "";
        if(str.contains("SPEED")) {
            stt = "<white>Speed";
            return stt;
        }
        if(str.contains("STRENGTH")) {
            stt = "<dark_red>Strength";
            return stt;
        }
        if(str.contains("HEALTH")) {
            stt = "<red>Health";
            return stt;
        }
        if(str.contains("DEFENSE")) {
            stt = "<dark_blue>Defense";
            return stt;
        }
        else {
            stt = "None";
            return stt;
        }
    }

    /**
     * set the attribute of an item
     * @param main a main instance
     * @param str the attribute
     * @param nb how many of the attribute
     * @param nbti the item
     * @return an item
     */
    @Deprecated
    public static  NBTItem setAttributes(Raxen main, @NotNull String str, Integer nb, NBTItem nbti) {
        switch (str) {
            case "SPEED":
                nbti.setInteger(NBT.SPEED, nb);
            case "STRENGTH":
                nbti.setInteger(NBT.STRENGTH, nb);
            case "HEALTH":
                nbti.setInteger(NBT.HEALTH, nb);
            case "DEFENSE":
                nbti.setInteger(NBT.DEFENSE, nb);
            case "LUCK":
                nbti.setInteger(NBT.LUCK, nb);
            case "MANA":
                nbti.setInteger(NBT.MANA, nb);
            default:
                error((Raxen) Bukkit.getPluginManager().getPlugin("Raxen"), "Cant set Attributes with an unknown attribute.");
        }
        return nbti;
    }

    /**
     * set the lore from the nbt
     * @param it an itemstack
     * @return the modified item
     */
    @Deprecated
    public static @NotNull ItemStack setLoreFromNBT(@NotNull ItemStack it) {
        ItemMeta itm = it.getItemMeta();
        List<Component> lore = new ArrayList<>();
        for (Component cp : itm.lore()) {
            lore.add(cp);
        }
        NBTItem nbti = new NBTItem(it);
        if(nbti.hasKey(NBT.SPEED)) {
            if(lore.size() > 2) {
                lore.add(2, colorComp(attributes("SPEED")).append(Component.text(nbti.getInteger(NBT.SPEED))));
            } else {
                lore.add(colorComp(attributes("SPEED")).append(Component.text(nbti.getInteger(NBT.SPEED))));
            }
        }
        if(nbti.hasKey(NBT.HEALTH)) {
            if(lore.size() > 1) {
                lore.add(1, colorComp(attributes("HEALTH")).append(Component.text(nbti.getInteger(NBT.HEALTH))));
            } else {
                lore.add(colorComp(attributes("HEALTH")).append(Component.text(nbti.getInteger(NBT.HEALTH))));
            }
        }
        if(nbti.hasKey(NBT.DEFENSE)) {
            if(lore.size() > 3) {
                lore.add(3, colorComp(attributes("DEFENSE")).append(Component.text(nbti.getInteger(NBT.DEFENSE))));
            } else {
                lore.add(colorComp(attributes("DEFENSE")).append(Component.text(nbti.getInteger(NBT.DEFENSE))));
            }
        }
        if(nbti.hasKey(NBT.STRENGTH)) {
            if(lore.size() > 3) {
                lore.add(3, colorComp(attributes("STRENGTH")).append(Component.text(nbti.getInteger(NBT.STRENGTH))));
            } else {
                lore.add(colorComp(attributes("STRENGTH")).append(Component.text(nbti.getInteger(NBT.STRENGTH))));
            }
        }
        if(nbti.hasKey(NBT.MANA)) {
            if(lore.size() > 3) {
                lore.add(3, colorComp(attributes("MANA")).append(Component.text(nbti.getInteger(NBT.MANA))));
            } else {
                lore.add(colorComp(attributes("MANA")).append(Component.text(nbti.getInteger(NBT.MANA))));
            }
        }
        if(nbti.hasKey(NBT.LUCK)) {
            if(lore.size() > 3) {
                lore.add(3, colorComp(attributes("LUCK")).append(Component.text(nbti.getInteger(NBT.LUCK))));
            } else {
                lore.add(colorComp(attributes("LUCK")).append(Component.text(nbti.getInteger(NBT.LUCK))));
            }
        }
        itm.lore(lore);
        it.setItemMeta(itm);
        return it;
    }

    public static ItemStack customModelData(ItemStack it, Integer in) {
        ItemMeta itm = it.getItemMeta();
        itm.setCustomModelData(in);
        it.setItemMeta(itm);
        return it;
    }
}
